#!/usr/bin/perl -w
#
# lookx - Resolve names and numbers using standard OS lookup mechanisms.
#
# Copyright (c) 2003-2011 Occam's Razor. All rights reserved.
#
# See the LICENSE file distributed with this code for restrictions on its use
# and further distribution.
# Original distribution available at <http://www.occam.com/tools/>.
#
# $Id: lookx.pl,v 5.1 2011/03/05 23:55:09 leonvs Exp $
#
# TODO
#

use strict;
use vars qw($opt_h $opt_V $opt_r $opt_m);
use Getopt::Std;
use Socket;

# global variables
my $progname;
my $mode;		# HOST, USER, GROUP, SERVICE
my $REV = 0;		# Run each answer through another lookup.
my $MACHINE_OUTPUT = 0;	# Make output easier to process by computer.
my @result = ();	# list of 2- or 3-tuples
my @questions = ();	# list of arguments, used for determining column width
my @answers = ();	# list of results, used for determining column width
my @revAnswers = ();	# list of reverse results, used for column width
my $lastQSeen = "";	# used to indent output w/ multiple results per arg
my $lastQASeen = "";	# used to indent output w/ multiple reverse results

processArgs();
foreach (@ARGV) {
	my $question = $_;
	push @questions, $question unless $MACHINE_OUTPUT;

	my @answer = lookup($question);
	foreach my $answer (@answer) {
		push @answers, $answer unless $MACHINE_OUTPUT;

		my @revAnswer = lookup($answer) if $REV;
		# The result of the first lookup may be bogus, in which case a
		# lookup returns an empty value. Check for that.
		if (@revAnswer) {
			foreach my $revAnswer (@revAnswer) {
				push @revAnswers, $revAnswer
					unless $MACHINE_OUTPUT;

				push @result,
					[ $question, $answer, $revAnswer ];
			}
		} else {
			push @result, [ $question, $answer ];
		}
	}
}
if ($MACHINE_OUTPUT) {
	foreach my $resultRef (@result) {
		print join "\t", @$resultRef;
		print "\n";
	}
} else {
	my ($longestQ) = reverse sort byLength @questions;
	my $qLen = length $longestQ;

	my ($longestA) = reverse sort byLength @answers;
	my $aLen = length $longestA;

	my ($longestRA) = reverse sort byLength @revAnswers if @revAnswers;
	my $raLen = length $longestRA if @revAnswers;
	
	foreach my $resultRef (@result) {
		# Just indent if we saw this question on previous line.
		my $question = (@$resultRef[0] eq $lastQSeen) ?
			"" : @$resultRef[0];
		$lastQSeen = @$resultRef[0];
		printf "%-${qLen}s", $question;

		# Just indent if we saw this q/a combo on previous line.
		my $answer = ("@$resultRef[0]:@$resultRef[1]" eq $lastQASeen) ?
			"" : @$resultRef[1];
		$lastQASeen = "@$resultRef[0]:@$resultRef[1]";
		printf " -> %-${aLen}s", $answer;

		printf " -> %-${raLen}s", @$resultRef[2] if @$resultRef[2];

		print "\n";
	}
}

sub processArgs {
	$progname = $0; $progname =~ s/.*\///;
	my $VERSION = "$progname 5.1\n";

	$mode = "HOST";		# Default is to look up hosts.
	$mode = "USER" if $progname =~ /^user/;
	$mode = "GROUP" if $progname =~ /^group/;
	$mode = "SERVICE" if $progname =~ /^serv/;

	my $USAGE = <<EOF;
$VERSION
usage:	$progname { -h | -V }
	$progname [-r] [-m] [name|number]...

	-h: Prints this usage statement and exits.
	-V: Prints version number and exits.
	-r: Performs both forward and reverse lookups for each argument.
	-m: Generates more machine-readable output (with tab delimiters).
	name|number: Space-separated list of names or numbers associated with
		hosts, users, groups, or services.
EOF

	getopts('hVrm') or die $USAGE;
	$opt_h and die $USAGE;
	$opt_V and die $VERSION;

	$opt_r and ++$REV;
	$opt_m and ++$MACHINE_OUTPUT;

	# Need at least one argument.
	@ARGV or die $USAGE;
}

# Returns a list value.
sub lookup {
	my $thingy = shift;

	return () if $thingy =~ /^(NO|INVALID|UNKNOWN)/;

	my $lookupFunc = "lookup$mode";
	no strict 'refs';
	return &$lookupFunc($thingy);
}

sub lookupHOST {
	my $thingy = shift;
	no strict 'subs';	# allow address type arg to gethostbyaddr()

	if ($thingy =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {	# IPv4
		my $result = inet_aton $thingy or return ("INVALID IP ADDRESS");
		$result = gethostbyaddr $result, AF_INET or
			return ("NO HOSTNAME FOUND");
		return ($result);
	} elsif ($thingy =~ /^[\da-fA-F]{0,4}([:.][\da-fA-F]{0,4})+$/) { # IPv6
		my $result = inet_aton $thingy or return ("INVALID IP ADDRESS");
		$result = gethostbyaddr $result, AF_INET6 or
			return ("NO HOSTNAME FOUND");
		return ($result);
	} elsif ($thingy =~ /^\w[\w.-]*$/) {				# name
		# A hostname should include only alphanumerics (including
		# underscores), hyphens, and dots, and begin with an
		# alphanumeric.
		my @results = gethostbyname $thingy or
			return ("NO IP ADDRESS FOUND");
		@results = map { inet_ntoa $_ } @results[4 .. $#results] or
			return ("UNKNOWN ERROR");
		return @results;
	} else {
		return ("INVALID HOSTNAME");
	}
}

sub lookupUSER {
	my $thingy = shift;

	if ($thingy =~ /^\d+$/) {					# UID
		my $result = getpwuid $thingy or return ("NO USER FOUND");
		return ($result);
	} elsif ($thingy =~ /^\w[\w\-\.]*$/) {				# name
		# A username should include only alphanumerics (including
		# underscores), hyphens, and dots, and begin with an
		# alphanumeric.
		my $result = getpwnam $thingy or return ("NO UID FOUND");
		return ($result);
	} else {
		return ("INVALID USERNAME");
	}
}

sub lookupGROUP {
	my $thingy = shift;

	if ($thingy =~ /^\d+$/) {					# GID
		my $result = getgrgid $thingy or return ("NO GROUP FOUND");
		return ($result);
	} elsif ($thingy =~ /^\w[\w\-\.]*$/) {				# name
		# A group name should include only alphanumerics (including
		# underscores), hyphens, and dots, and begin with an
		# alphanumeric.
		my $result = getgrnam $thingy or return ("NO GID FOUND");
		return ($result);
	} else {
		return ("INVALID GROUPNAME");
	}
}

sub lookupSERVICE {
	my $thingy = shift;

	if ($thingy =~ /^(UDP|TCP) \d+$/) {				# port
		my ($proto, $port) = split / /, $thingy;
		my $result = getservbyport $port, lc $proto or
			return ("NO $proto SERVICE FOUND");
		$result = "$proto $result";
		return ($result);
	} elsif ($thingy =~ /^(UDP|TCP) \w[\w\-\+\.]*$/) {		# name
		# A service name should include only alphanumerics (including
		# underscores), hyphens, pluses, and dots, and begin with an
		# alphanumeric.
		my ($proto, $svc) = split / /, $thingy;
		my $result = getservbyname $svc, lc $proto or
			return ("NO $proto PORT FOUND");
		$result = "$proto $result";
		return ($result);
	} elsif ($thingy =~ /^\d+$/ || $thingy =~ /^\w[\w\-\+\.]*$/) {
		# No protocol specified.
		my @results = ();
		push @results, lookup("TCP $thingy");
		push @results, lookup("UDP $thingy");
		return @results;
	} else {
		return ("INVALID SERVICE");
	}
}

sub byLength {
	length $a <=> length $b;
}
